from __future__ import annotations

from enum import StrEnum

from pydantic import Field

from shipaw.models.base import ShipawBaseModel
from shipaw.models.ship_types import ShipDirection
from shipaw.providers.provider_abc import ProviderName


class AlertType(StrEnum):
    ERROR = 'ERROR'
    WARNING = 'WARNING'
    NOTIFICATION = 'NOTIFICATION'


class Alert(ShipawBaseModel):
    code: int | None = None
    message: str
    type: AlertType = AlertType.NOTIFICATION

    def __eq__(self, other):
        if not isinstance(other, Alert):
            return NotImplemented
        return (self.code, self.message, self.type) == (other.code, other.message, other.type)

    def __hash__(self):
        return hash((self.code, self.message, self.type))

    @classmethod
    def from_exception(cls, e: Exception):
        return cls(message=str(e), type=AlertType.ERROR)


class Alerts(ShipawBaseModel):
    alert: list[Alert] = Field(default_factory=list[Alert])

    @property
    def errors(self):
        return [a for a in self.alert if a.type == AlertType.ERROR]

    @property
    def warnings(self):
        return [a for a in self.alert if a.type == AlertType.WARNING]

    @property
    def notifications(self):
        return [a for a in self.alert if a.type == AlertType.NOTIFICATION]

    def __bool__(self):
        return bool(self.alert)

    def __add__(self, other: Alerts | Alert):
        if not isinstance(other, Alerts) and not isinstance(other, Alert):
            raise TypeError(f'Expected Alerts or Alert instance, got {type(other)}')
        if isinstance(other, Alert):
            other = Alerts(alert=[other])
        combined = set(self.alert) | set(other.alert)
        return Alerts(alert=list(combined))

    def __iadd__(self, other: Alerts | Alert):
        if not isinstance(other, Alerts) and not isinstance(other, Alert):
            raise TypeError(f'Expected Alerts or Alert instance, got {type(other)}')
        if isinstance(other, Alert):
            other = Alerts(alert=[other])
        self.alert = list(set(self.alert) | set(other.alert))
        return self

    def __sub__(self, other: Alerts | Alert):
        if not isinstance(other, Alerts) and not isinstance(other, Alert):
            raise TypeError(f'Expected Alerts or Alert instance, got {type(other)}')
        if isinstance(other, Alert):
            other = Alerts(alert=[other])
        diff = set(self.alert) - set(other.alert)
        return Alerts(alert=list(diff))

    def __contains__(self, alert: Alert):
        return alert in set(self.alert)

    @classmethod
    def empty(cls):
        return cls(alert=[])


async def maybe_alert_phone_number(phone_num: str):
    """Alert if phone number is not 11 digits or does not start with 01, 02 or 07"""
    alerts = Alerts.empty()
    if len(phone_num) != 11 or not phone_num[0:2] == '07':
        alerts += Alert(
            type=AlertType.ERROR,
            message=f'The Mobile phone number ({phone_num}) must be 11 digits and begin with 07. Unable to send with no phone number (try "Use Home Base Mobile").',
        )
    return alerts


async def check_royal_mail(shipment_request) -> Alerts:
    alerts = Alerts.empty()
    if shipment_request.provider_name == ProviderName.ROYAL_MAIL:
        if shipment_request.shipment.direction != ShipDirection.OUTBOUND:
            msg = f'Royal Mail only supports OUTBOUND shipments at this time. Please use another provider for {shipment_request.shipment.direction} shipments.'
            alerts += Alert(message=msg, type=AlertType.ERROR)
    return alerts



from pydantic import EmailStr, conlist, constr, field_validator, model_validator

from shipaw.models.base import ShipawBaseModel


class Contact(ShipawBaseModel):
    contact_name: str
    email_address: str
    mobile_phone: str
    phone_number: str | None = None

    @model_validator(mode='after')
    def phone_is_none(self):
        if not self.phone_number:
            self.phone_number = self.mobile_phone
        return self

    @field_validator('mobile_phone', mode='after')
    def clean_mobile_phone(cls, v):
        return v.replace(' ', '').replace('-', '')


class Address(ShipawBaseModel):
    business_name: str
    address_lines: list[str] = conlist(item_type=str, max_length=3, min_length=1)
    town: constr(max_length=25)
    postcode: constr(max_length=16)
    county: str | None = None
    country: str = 'GB'

    def get_address_lines_dict(self, prefix='address_line') -> dict[str, str]:
        return {f'{prefix}{i + 1}': line for i, line in enumerate(self.address_lines) if line}

    @field_validator('address_lines', mode='after')
    def check_address_lines(cls, v):
        if not v or len(v) == 0:
            raise ValueError('At least one address line is required')
        if len(v) > 3:
            v[2] = ', '.join(v[2:])
        if len(v) < 3:
            v += [''] * (3 - len(v))
        return v[:3]


class LongContact(ShipawBaseModel):
    contact_name: str
    email_address: str | EmailStr
    mobile_phone: str
    phone_number: str | None = None

    business_name: str
    address_lines: list[str] = conlist(item_type=str, max_length=3, min_length=1)
    town: constr(max_length=25)
    postcode: constr(max_length=16)
    country: str = 'GB'

    @model_validator(mode='after')
    def phone_is_none(self):
        if not self.phone_number:
            self.phone_number = self.mobile_phone
        return self

    def get_address_lines_dict(self, prefix='address_line') -> dict[str, str]:
        return {f'{prefix}{i + 1}': line for i, line in enumerate(self.address_lines) if line}

    @field_validator('address_lines', mode='after')
    def check_address_lines(cls, v):
        if not v or len(v) == 0:
            raise ValueError('At least one address line is required')
        if len(v) > 3:
            v[2] = ', '.join(v[2:])
        if len(v) < 3:
            v += [''] * (3 - len(v))
        return v[:3]


class FullContact(ShipawBaseModel):
    contact: Contact
    address: Address


class AddressChoice(ShipawBaseModel):
    address: Address
    score: int



from enum import StrEnum


def enum_as_tups(enum_type: type[StrEnum]) -> list[tuple[str, str]]:
    return list(enum_type.__members__.items())


def enum_as_dict(enum_type: type[StrEnum]) -> dict[str, str]:
    return dict(enum_type.__members__.items())


def enum_lookup(*, enum_type: type[StrEnum], attr_name: str) -> str:
    res = getattr(enum_type, attr_name, None)
    if not res:
        raise ValueError(f'Invalid attr_name: {attr_name}')
    return res


def enum_reverse_lookup(*, enum_type: type[StrEnum], attr_value: str) -> str:
    res = next(name for name in enum_type.__dict__.values() if name == attr_value)
    if not res:
        raise ValueError(f'Invalid service code: {attr_value}')
    return res



from pprint import pformat

from fastapi import APIRouter, Body, Depends
from loguru import logger
from starlette.requests import Request
from starlette.responses import JSONResponse

from shipaw.config import SHIPAW_SETTINGS
from shipaw.fapi.ui_funcs import make_nice_str
from shipaw.fapi.alerts import Alerts
from shipaw.fapi.backend import (
    errored_shipment,
    maybe_alert_apc,
    notify_dev,
    notify_version,
    try_book_shipment,
    try_get_write_label,
)
from shipaw.fapi.form_data import provider_from_form, shipment_request_form, shipment_request_form_json
from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import ShipawTemplate, ShipawTemplateResponse
from shipaw.logging import log_obj
from shipaw.models.shipment import Shipment
from shipaw.providers.registry import PROVIDER_REGISTER
from urllib.parse import unquote
from royal_mail_combined.parcels_apis.address.models import AddressRecordDef, AddressSummaryDef

router = APIRouter()


# router.mount('/static', StaticFiles(directory=str(SHIPAW_SETTINGS.static_dir)), name='static')


@router.post('/shipping_form', response_model=ShipawTemplateResponse)
async def shipping_form_api(request: Request, shipment: Shipment = Body(...)) -> ShipawTemplateResponse:
    log_obj(shipment, 'Shipment received at /ship_form:')

    alerts: Alerts = request.app.alerts
    alerts += notify_dev()
    alerts += await notify_version(request)

    tmplt = ShipawTemplate(template_path='shipping_form_container.html', context={'shipment': shipment})
    return ShipawTemplateResponse(template=tmplt, alerts=alerts)


@router.post('/order_summary', response_model=ShipawTemplateResponse)
async def order_summary_api(
        request: Request,
        shipment_request: ShipmentRequest = Depends(shipment_request_form),
) -> ShipawTemplateResponse:
    log_obj(shipment_request, 'ShipmentRequest received at shipaw/order_summary:')
    context = {'shipment_request': shipment_request}

    # check phone number
    # alerts = await maybe_alert_phone_number(shipment_request.shipment.remote_full_contact.contact.mobile_phone)
    # check not apc + dropoff
    alerts = await maybe_alert_apc(shipment_request)
    return ShipawTemplateResponse(
        template=ShipawTemplate(template_path='/order_summary.html', context=context),
        alerts=alerts,
    )


@router.post('/order_results', response_model=ShipawTemplateResponse)
async def order_results_api(
        request: Request,
        shipment_request: ShipmentRequest = Depends(shipment_request_form_json),
) -> ShipawTemplateResponse:
    shipment_response = await try_book_shipment(shipment_request)
    # await try_get_write_label(shipment_request, shipment_response)

    if shipment_response.alerts.errors:
        return await errored_shipment(shipment_response)

    log_obj(shipment_response, 'Shipment Booked')
    await try_get_write_label(shipment_request, shipment_response)

    if hasattr(request.app, 'callback'):
        await request.app.callback(shipment_request, shipment_response)

    shipment_response.template = ShipawTemplate(
        template_path='/order_results.html',
        context={'shipment_request': shipment_request, 'response': shipment_response},
    )
    return ShipawTemplateResponse.model_validate(shipment_response, from_attributes=True)


@router.get('/providers', response_class=JSONResponse)
async def providers():
    dflt = SHIPAW_SETTINGS.default_provider_name
    available = sorted(PROVIDER_REGISTER.keys(), key=lambda p: p != dflt)
    provider_response = {make_nice_str(_): _ for _ in available}
    return JSONResponse(provider_response)


@router.get('/provider_directions/{provider_name}', response_class=JSONResponse)
async def provider_directions(provider_name: str):
    provider = await provider_from_form(provider_name)
    directions = provider.valid_directions
    dir_dict = {make_nice_str(_.value): _.value for _ in directions}
    return JSONResponse(dir_dict)


@router.get('/provider_direction_services/{provider_name}/{direction}', response_class=JSONResponse)
async def provider_direction_services(provider_name: str, direction: str):
    provider = await provider_from_form(provider_name)
    dflt = provider.default_service
    available = sorted(provider.valid_direction_services.get(direction, []), key=lambda s: s.value != dflt)
    res_dir = {make_nice_str(_.name): _.value for _ in available}
    return JSONResponse(res_dir)


@router.get('/address_search/{search_text}', response_model=list[AddressSummaryDef])
async def address_search(search_text: str):
    provider = PROVIDER_REGISTER.get('ROYAL_MAIL')
    res = provider.client.address_search(search_text)
    # addresses = [_.model_dump(mode='json', by_alias=True) for _ in res.addresses]
    addresses = [_.model_dump(mode='json', by_alias=True) for _ in res.addresses]
    addr_string = pformat(res, indent=2)
    logger.debug(f'Address search for "{search_text}" returned:\n{addr_string}')
    return res.addresses


@router.get('/address_retrieve/{addr_id}', response_model=AddressRecordDef)
async def address_retrieve(addr_id: str):
    addr_id = unquote(addr_id)
    provider = PROVIDER_REGISTER.get('ROYAL_MAIL')
    res = provider.client.address_retrieve(addr_id)
    addr_string = pformat(res, indent=2)
    logger.debug(f'Address search for "{addr_id}" returned:\n{addr_string}')
    return res



import sqlalchemy
from sqlalchemy import Column
from sqlmodel import Field, SQLModel

from shipaw.fapi.requests import ShipmentRequest
from shipaw.models.address import Address, Contact, FullContact
from shipaw.models.shipment import Shipment
from shipaw.sql_helpers import optional_json_field, required_json_field


class TableModel(SQLModel):
    """Base class for SQLModel tables."""

    id: int | None = Field(default=None, primary_key=True)


class AddressTable(Address, TableModel, table=True):
    __tablename__ = 'address'
    # address_lines: list[str] = required_json_field()
    address_lines: list[str] = Field(..., sa_column=Column(sqlalchemy.JSON))


class ContactTable(Contact, TableModel, table=True):
    __tablename__ = 'contact'


class FullContactTable(FullContact, TableModel, table=True):
    __tablename__ = 'full_contact'
    contact: Contact = required_json_field(Contact)
    address: Address = required_json_field(Address)


class ShipmentTable(Shipment, TableModel, table=True):
    __tablename__ = 'shipment'
    recipient: FullContact = required_json_field(FullContact)
    sender: FullContact | None = optional_json_field(FullContact)
    context: dict = Field(default_factory=dict, sa_column=Column(sqlalchemy.JSON))


class ShipmentRequestTable(ShipmentRequest, TableModel, table=True):
    """SQLModel table for ShipmentRequest"""

    __tablename__ = 'shipment_request'
    shipment: Shipment = required_json_field(Shipment)



from loguru import logger

from shipaw.providers.provider_abc import ShippingProvider

PROVIDER_TYPE_REGISTER: dict[str, type[ShippingProvider]] = {}
PROVIDER_REGISTER: dict[str, ShippingProvider] = {}


def register_provider_type(cls):
    PROVIDER_TYPE_REGISTER[cls.name] = cls
    logger.info(f'Registered provider type {cls.name}')
    return cls


def register_provider_instance(instance: ShippingProvider):
    PROVIDER_REGISTER[instance.name] = instance
    logger.info(f'Registered active provider instance {instance.name}')
    return instance



from apc_hypaship.models.response.resp import BookingResponse

from shipaw.fapi.alerts import Alerts, Alert
from shipaw.fapi.responses import ShipmentResponse
from shipaw.models.shipment import Shipment as ShipmentAgnost


def errored_booking(shipment: ShipmentAgnost, response: BookingResponse):
    messages = response.orders.order.messages
    fieldname, message = strip_apc_error_msgs(messages)
    return errored_response(fieldname, message, shipment, response.model_dump())


def strip_apc_error_msgs(messages):
    fieldname = messages.error_fields.error_field.field_name
    message = messages.error_fields.error_field.error_message
    return fieldname, message


def errored_response(fieldname: str, message: str, shipment: ShipmentAgnost, data: dict):
    alerts = Alerts(alert=[Alert(message=f'Error booking shipment: {fieldname}: {message}')])
    return ShipmentResponse(
        alerts=alerts,
        shipment=shipment,
        shipment_num='FAILED TO BOOK',
        tracking_link='FAILED TO BOOK',
        data=data,
        success=False,
    )



from shipaw.models.address import Address, Contact
from shipaw.models.base import ShipawBaseModel
from shipaw.models.shipment import Shipment
from shipaw.providers.provider_abc import ProviderName, ShippingProvider
from shipaw.providers.registry import PROVIDER_REGISTER


# class Authentication(ShipawBaseModel):
#     # todo SecretStr!!!!
#     user_name: Annotated[str, StringConstraints(max_length=80)]
#     password: Annotated[str, StringConstraints(max_length=80)]


class ShipmentRequest(ShipawBaseModel):
    # id: uuid.UUID = uuid.uuid4()
    shipment: Shipment
    provider_name: ProviderName
    service_code: str

    @property
    def provider(self) -> ShippingProvider:
        if not self.provider_name:
            raise ValueError('Provider name is not set')
        if self.provider_name not in PROVIDER_REGISTER:
            raise ValueError(f'Unknown provider: {self.provider_name}')
        return PROVIDER_REGISTER[self.provider_name]


class AddressRequest(ShipawBaseModel):
    postcode: str
    address: Address | None = None
    contact: Contact | None = None



import base64
from pprint import pformat
from typing import ClassVar, override

from loguru import logger
from royal_mail_combined import RoyalMailClient
from royal_mail_combined.click_and_drop_api.models import (
    CreateOrderRequest,
    CreateOrdersRequest,
    CreateOrdersResponse,
)
from royal_mail_combined.click_and_drop_api.models.return_models import ReturnRequestContainer, ReturnResponseContainer
from royal_mail_combined.config import RoyalMailSettingsGlobal
from royal_mail_combined.converters_no_import import tracking_link
from royal_mail_combined.core.consts_types import RoyalMailServiceCodes

from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import ShipmentResponse
from shipaw.models.ship_types import ShipDirection
from shipaw.models.shipment import Shipment
from shipaw.providers.provider_abc import ProviderName, ShippingProvider
from shipaw.providers.registry import register_provider_type
from shipaw.providers.royal_mail.royal_mail_funcs import (
    full_contact_from_rm,
    inbound_shipment_from_agnostic,
    outbound_shipment_from_agnostic,
)
import io
from pypdf import PdfWriter, PdfReader


def merge_pdf_bytes(pdf_bytes_list: list[bytes]) -> bytes:
    writer = PdfWriter()
    for pdf_bytes in pdf_bytes_list:
        reader = PdfReader(io.BytesIO(pdf_bytes))
        for page in reader.pages:
            writer.add_page(page)
    output = io.BytesIO()
    writer.write(output)
    return output.getvalue()


def build_booking_response_inbound(rm_response: ReturnResponseContainer, shipment: Shipment):
    unique_item_ids_str = ';'.join([order.shipment.unique_item_id for order in rm_response.created_orders])
    tracking_nums = [order.shipment.tracking_number for order in rm_response.created_orders]
    tracking_links = [tracking_link(_) for _ in tracking_nums]
    tracking_links_str = ';'.join(tracking_links)

    # combine labels for multiple orders
    labels_bytes = [base64.b64decode(order.label) for order in rm_response.created_orders]
    combined_pdf = merge_pdf_bytes(labels_bytes)

    return ShipmentResponse(
        tracking_links=tracking_links,
        shipment_numbers=tracking_nums,
        label_data=combined_pdf,
        shipment=shipment,
        shipment_num=unique_item_ids_str,
        tracking_link=tracking_links_str,
        data=rm_response.model_dump(),
        status='Success',
        success=True,
    )


def print_response_errors(rm_response: CreateOrdersResponse):
    logger.error(f'Errors booking outbound shipment: {pformat(rm_response.failed_orders, indent=2, width=120)}')


@register_provider_type
class RoyalMailProvider(ShippingProvider):
    settings: RoyalMailSettingsGlobal

    name: ClassVar[ProviderName] = ProviderName.ROYAL_MAIL
    settings_type: ClassVar[type[RoyalMailSettingsGlobal]] = RoyalMailSettingsGlobal
    service_codes_type: ClassVar[type[RoyalMailServiceCodes]] = RoyalMailServiceCodes
    default_service: ClassVar[RoyalMailServiceCodes] = RoyalMailServiceCodes.EXPRESS_24
    valid_directions: ClassVar[list[ShipDirection]] = [
        ShipDirection.OUTBOUND,
        ShipDirection.INBOUND,
        ShipDirection.DROPOFF,
    ]

    valid_direction_services = {
        ShipDirection.OUTBOUND: [RoyalMailServiceCodes.TRACKED_24, RoyalMailServiceCodes.EXPRESS_24],
        ShipDirection.INBOUND: [RoyalMailServiceCodes.TRACKED_24_RTN],
        ShipDirection.DROPOFF: [RoyalMailServiceCodes.TRACKED_24_RTN],
    }

    _client: RoyalMailClient | None = None

    @override
    def is_sandbox(self) -> bool:
        # Royal mail do not have a test environment. that's fun
        return False

    @property
    @override
    def client(self) -> RoyalMailClient:
        if self._client is None:
            if self.settings is None:
                raise ValueError('Settings must be set before using the client')
            self._client = RoyalMailClient(settings=self.settings)
        return self._client

    @classmethod
    @override
    def agnostic_shipment(cls, orders_req: CreateOrdersRequest) -> Shipment:
        order = orders_req.items[0]
        return Shipment(
            recipient=full_contact_from_rm(order.recipient),
            boxes=len(order.packages),
            shipping_date=order.planned_despatch_date.date(),
            direction=ShipDirection.OUTBOUND,
            reference=order.order_reference,
        )

    @override
    def provider_shipment(
        self, shipment: Shipment, service_code: RoyalMailServiceCodes
    ) -> CreateOrderRequest | ReturnRequestContainer:
        provider_service = self.service_codes_type(service_code)

        if shipment.direction == ShipDirection.OUTBOUND:
            return outbound_shipment_from_agnostic(shipment, provider_service)
        else:
            return inbound_shipment_from_agnostic(shipment, provider_service)

    @override
    def book_shipment_agnostic(self, shipment_request: ShipmentRequest) -> ShipmentResponse:
        shipment = shipment_request.shipment
        service = self.service_codes_type(shipment_request.service_code)

        resp = None
        match shipment_request.shipment.direction:
            case ShipDirection.OUTBOUND:
                order_create = outbound_shipment_from_agnostic(shipment, service)
                resp = self.client.book_outbound_shipment(order_create)
                return self.build_booking_response_outbound(resp, shipment)

            case ShipDirection.DROPOFF:
                return_req_container = inbound_shipment_from_agnostic(shipment, service)
                resp = self.client.book_inbound_shipment(return_req_container)
                return build_booking_response_inbound(resp, shipment)

            case ShipDirection.INBOUND:
                return_req_container = inbound_shipment_from_agnostic(shipment, service)
                resp = self.client.book_inbound_shipment_with_collection(
                    return_req_container, collection_date=shipment.shipping_date
                )
                return build_booking_response_inbound(resp, shipment)

            case _:
                raise ValueError(f'Invalid shipment direction {shipment.direction} for booking')

    def build_booking_response_outbound(self, rm_response: CreateOrdersResponse, shipment: Shipment):
        if rm_response.errors_count > 0:
            print_response_errors(rm_response)

        label_data = self.client.get_label_data(rm_response.success_idents_str)
        fetched = self.client.fetch_specific_orders(rm_response.success_idents_str)

        success = rm_response.errors_count == 0
        tracking_numbers = [order.tracking_number for order in fetched]
        tracking_links = [tracking_link(_) for _ in tracking_numbers]
        res = ShipmentResponse(
            label_data=label_data,
            shipment=shipment,
            shipment_num=rm_response.success_idents_str,
            tracking_link=', '.join(tracking_links),
            data={_.order_identifier: _ for _ in fetched},
            status='Success' if success else 'FAIL',
            success=success,
        )
        return res

    @override
    def fetch_label_content(self, shipment_num: str) -> bytes:
        return self.client.get_label_data(shipment_num)
